C *********************************************************************
C       "A data-driven method for modelling and reconstructing
C	     social interactions in moving animal groups"
C	  R. Escobedo, V. Lecheval, V. Papaspyros, F. Bonnet,
C		  F. Mondada, C. Sire, G. Theraulaz
C
C 				questions/comments: escobedor@gmail.com
C ------- PREAMBLE ----------------------------------------------------
C Fortran code to extract the interaction functions in simple cases:
C	From data stored in a 2D-matrix DPhi of size NdxNpsi, the code
C	extracts the 1D functions fR and gR of size Nd and Npsi (resp.)
C	such that
C		      fR(id) x gR(ipsi) = DPhi(id,ipsi).
C
C	1. From data generated with analytical expressions:
C	   1.1 data homogeneously distributed (eps = 1),
C	   1.2 using the data distribution of real fish experiments,
C	       borrowed from the experimental datafiles used in the
C	       article and in [Calovi2018].
C
C	2. From real data of fish experiments:
C	   D. rerio: 2zebraf.dat, H. rhodostomus: 2hemigr.dat
C	   Datafiles have 5 columns structured in events:
C		  event_id    fish_id  time   x(t)  y(t)
C		1 --> Nevent   1 or 2  (sec)  (m)   (m)
C
C	To select the option, edit this file and, in lines 93--95,
C	choose the corresponding value of IOP:
C		IOP = 1: case 1.1: analytical & homogeneous 
C		IOP = 2: case 1.2: analytical & real distribution
C		IOP = 3: case 2:   from experimental data
C
C	Several analytical expressions are proposed in FUNCTIONS
C	fR and gR, with adjustable levels of noise, and two examples
C	of non-separable functions.
C	See Sec. 3.2 of XTR subroutine (around line 400 of this file).
C
C - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C Structure of the code:
C	xtract.f: (this file) contains the main program, which calls
C		a subroutine "xtr" and two functions fRepu, gRepu,
C		to generate data from analytical expressions.
C		xtract.f reads the datafile and define the dimensions
C		of the variables used in the subroutine xtr.
C
C	   xtr:	(in this file) subroutine performing the extraction.
C		Details are given in the preamble of the subroutine.
C
C        iters: (in this file) subroutine performing the iteratve
C		process. INPUTS: the values of the heading change
C		in a matrix PP of size (Nd,Npsi). OUTPUTS: vectors
C		f(Nd) and g(Npsi) to which iterations converge, in
C		two datafile 02.fR-end.dat, 02.gR-end.dat, and the
C		evolution of iterations (in 3D datafiles).
C
C - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C The code generates six datafiles, three for each function fR and gR:
C
C	00.fR-analytic.dat: 2D-dataset of the analytical expression
C			    (using 101 nodes).
C
C	01.fR-iters.dat: 3D-dataset of the evolution of the iterative
C			 process (each iteration contains Nd values).
C						
C	02.fR-end.dat: 2D-dataset of the final state to which the
C		       iterative process has converged (Nd values).
C
C	The corresponding files are produced for gR(Npsi).
C
C Other files:
C
C	pl-2f.gnuplot:  external file to be used in a xterminal with
C			the (freely available) software "gnuplot".
C
C	2zebraf.dat, 2hemigr.dat: datafiles from real experiments.
C
C - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
C We recomend to copy all these files in the same folder. Then:
C 	Compile with:
C 		> gfortran -O2 xtract.f -o dale
C 	Run with:
C 		> dale
C	Represent the result (i.e., the interaction functions) with:
C		> gnuplot
C		gnuplot> load "pl-2f.gnuplot"
C
C -------- END OF PREAMBLE --------------------------------------------
C *********************************************************************
C
	PROGRAM Xtract
	PARAMETER(Nfish=2, NevMax=3000) !NevMax: Max number of events
	CHARACTER*30 filename
	REAL*8 RArena, BL
	INTEGER Nt(NevMax),Nle(NevMax)
C
C Select option (see preamble)
	IOP = 1 ! analytical & homogeneous 
c	IOP = 2 ! analytical & real distribution
c	IOP = 3 ! from experimental data
C
C Hemigrammus: comment/uncomment filename to select datafile:
	filename = '2hemigr.dat'
C Zebrafish:
	filename = '2zebraf.dat'
C
C Preliminary analysis of data file: File structure: five columns
C	  event_id    fish_id  time   x(t)  y(t)
C	1 --> Nevent   1 or 2  (sec)  (m)   (m)
C
C Lines:
C    Nle(iev): vector of the number of data of event "iev" (all fish)
C    Nt = Nle/Nfish: vector of the number of times per fish of event iev
C    Nlines: number of lines ( i.e. times) of the original datafile
C    Ntmax: number of times of the longest event ( = max(Nt) ).
C Count number of events and total number of lines:
	Nlines = 0
	OPEN(2,file=filename)
	DO it=1,10000000
	  READ(2,*,end=7) Nevent
	  Nlines = Nlines + 1
	ENDDO
7	CLOSE(2)
C Count number of lines per event:
	OPEN(2, FILE=filename)
	ievent = 1
	Nle = 0 !vector
	 Nt = 0 !vector
	DO i=1,Nlines
	  READ(2,*,END=8) ievt
	  IF(ievt.EQ.ievent) Nle(ievent) = Nle(ievent) + 1
	  IF(ievt.GT.ievent) THEN
	    ievent = ievent + 1
	    Nle(ievent) = Nle(ievent) + 1
	  ENDIF
	ENDDO
8	CLOSE(2)
	Nt = Nle/Nfish !vector
C
C Numer of lines of largest event:
	Ntmax = maxval(Nt)	!f90
C
C ---------------------------------------------------------------------
C Number of boxes of the discretization of each state variable:
C (this step is done here to adjust easily the dimension of variables)
	  Nd = 21
	Npsi = 61
C Call main part of the program:
	CALL XTR(Nevent,Nfish,Nlines,Nle,Nt,Ntmax,filename,Nd,Npsi,IOP)
	END

C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C SUBROUTINE CONTAINING THE MAIN PROGRAM
C Structure of the subroutine:
C	1) Read the data from datafile. This is necessary to know the
C	   distribution of data (i.e. the number of data in each box).
C	2) Calculate observables from data of real fish experiments.
C	   This is necessary to calculate the vector DPhi, which is
C	   the heading change, and its average in each box.
C	3) Perform the iterative process and represent the solution.
C	4) Represent the analytical functinos (when possible).
C *********************************************************************
C	
	SUBROUTINE XTR(Nevent,Nfish,Nlines,Nle,Nt,Ntmax,filename
     &							  ,Nd,Npsi,IOP)
	CHARACTER*30 filename
	INTEGER Nle(Nevent),Nt(Nevent)
	REAL*8  Deltat(Nevent,Nfish,Ntmax), RArena, BL, xm
	REAL*8      xx(Nevent,Nfish,Ntmax), x(Ntmax)
	REAL*8      yy(Nevent,Nfish,Ntmax), y(Ntmax)
	REAL*8       r(Nevent,Nfish,Ntmax)
	REAL*8      Vx(Nevent,Nfish,Ntmax)
	REAL*8      Vy(Nevent,Nfish,Ntmax)
	REAL*8       V(Nevent,Nfish,Ntmax)
	REAL*8     Phi(Nevent,Nfish,Ntmax), PhiP1, PI
	REAL*8    DPhi(Nevent,Nfish,Ntmax)
	REAL*8     dij(Nevent,Nfish,Ntmax,Nfish), dd(Nfish)
	REAL*8 Thetaij(Nevent,Nfish,Ntmax,Nfish)
	REAL*8   psiij(Nevent,Nfish,Ntmax,Nfish)
	REAL*8  q,q1,q2,q3,q4,q5, qx,qy !auxiliary variables
C
	REAL*8   PP(Nd,Npsi), dijmin,dijmax, psiijmin,psiijmax
	INTEGER eps(Nd,Npsi),epstot
	REAL*8 fR(Nd),gR(Npsi), fRepu,gRepu
C
	PI = DACOS(-1.D0)
	CALL SRAND(-94187951)
C
C Hemigrammus:
	IF(filename.EQ.'2hemigr.dat') THEN
	  BL = 0.03D0
	  RArena = 0.250D0
	ENDIF
C Zebrafish:
	IF(filename.EQ.'2zebraf.dat') THEN
	  BL = 0.045D0
	  RArena = 0.29D0
	ENDIF
c Range of state variables for reconstruction:
	  dijmin = 0.D0
	  dijmax = 1.2D0*RArena
	psiijmin = - PI + 1.D-8
	psiijmax = - psiijmin
C ------------------------- Summary of datafile: ----------------------
	WRITE(*,*) ' filename: ',filename
	IF(IOP.EQ.1) WRITE(*,*) 'OPTION: analytical & homogeneous '
	IF(IOP.EQ.2) WRITE(*,*) 'OPTION: analytical & real distr.'
	IF(IOP.EQ.3) WRITE(*,*) 'OPTION: from experimental data'
	WRITE(*,'(A8,1x,I2)')   '  #fish:',Nfish
	WRITE(*,'(A8,1x,F5.3)') '  BL(m):',BL
	WRITE(*,'(A8,1x,F5.3)') '  Arena:',RArena
	WRITE(*,'(A14,1x,I8)')  ' total #lines:',Nlines
	WRITE(*,*) '#lines per event (NOT SHOWN)'
	WRITE(*,*) '& time length (= #lines/Nfish)'
	Nltot = 0
	DO iev=1,Nevent
c	  WRITE(*,'(I3,2x,I6,2x,I6)') iev,Nle(iev),Nt(iev)
	  Nltot = Nltot + Nle(iev)
	ENDDO
	WRITE(*,'(A17,I4,2x,I7)') '#events & lines: ',Nevent,Nlines
	WRITE(*,'(A22,I7)') '#lines longest event: ',Ntmax
	WRITE(*,*)
C ---------------------------- End of summary -------------------------
C
C ***************************
C 1) Read data from datafile:
C ***************************
C Initialize time intervals between kicks:
	Deltat = 0.5D0 !vector
C Read ALL DATA:
	OPEN(2,FILE=filename)
	DO ievent = 1,Nevent
C -- Read the data of the "Nfish" fish of the "ievent-th" event:
	  DO ifish = 1,Nfish
	    DO it=1,Nt(ievent)
	      READ(2,*) ievt, ipd, Deltat(ievent,ifish,it)
     &		 ,xx(ievent,ifish,it),yy(ievent,ifish,it)
C
	      IF(ievt.NE.ievent.OR.ipd.NE.ifish) THEN
		WRITE(*,*) '----> problem:'
		WRITE(*,*) it,'event: ',ievt,ievent
		WRITE(*,*)' or '
		WRITE(*,*) it,'  ifish: ',ipd,ifish
		WRITE(*,*) 'Nt(ievent): ',Nt(ievent)
		WRITE(*,*) 'Ctrl-C, or enter 0 to continue'
		READ(*,*) q
	      ENDIF
	    ENDDO
	  ENDDO
	  WRITE(*,9999,ADVANCE='NO') CHAR(13), ievent,'-th event read'
	ENDDO
	CLOSE(2)
	WRITE(*,*) '--> Entire file read'
C
C Build the vector Deltat of time intervals:
C (what was read from 'filename' were instants of time, not durations)
	DO iev = 1,Nevent
	  DO ifish = 1,Nfish
	    DO it=1,Nt(iev)-1
	      Deltat(iev,ifish,it) = 
     &	      Deltat(iev,ifish,it+1) - Deltat(iev,ifish,it)
	    ENDDO
	    Deltat(iev,ifish,Nt(iev)) = 0.5D0 ! son segundos
	  ENDDO
	ENDDO
C 
C *********************************************
C 2) Calculate state variables and observables:
C  - individual: radial position, speed, heading angle, heading change
C  - collective: distance between fish, angle of perception
C *********************************************
	WRITE(*,*) 'Calculate quantities: r, V, Phi, d_ij, psi_ij...'
	DO iev = 1,Nevent
          !Remove 1 seconds from the initial and final transients:
	  itini = 1
	  itend = Nt(iev)-1 !-1 and +1 to have it-1 for Vx
C
	  !Individual behavior observables:
	  DO ifish = 1,Nfish
	    x = 0.D0 !vector
	    y = 0.D0 !vector
	    DO it = itini,itend
	      x(it) = xx(iev,ifish,it) ! x, y are auxiliary variables
	      y(it) = yy(iev,ifish,it) ! used to lighten notation
	    ENDDO  
	    DO it = itini,itend
	    !Radial position, speed, acceleration... etc.
	       r(iev,ifish,it) = DSQRT(x(it)**2 + y(it)**2)
	      Vx(iev,ifish,it) = (x(it) - x(it-1))/Deltat(iev,ifish,it-1)
	      Vy(iev,ifish,it) = (y(it) - y(it-1))/Deltat(iev,ifish,it-1)
	       V(iev,ifish,it) = DSQRT(Vx(iev,ifish,it)**2
     &				    + Vy(iev,ifish,it)**2)
	    !Angle PHI of velocity vector (= HEADING), at it and it+1:
	      q1 = ATAN2(y(it)   - y(it-1),x(it)   - x(it-1))
	      q2 = ATAN2(y(it+1) - y(it)  ,x(it+1) - x(it))
	      Phi(iev,ifish,it) = q1
	      PhiP1             = q2
	    !Heading change dphi: diff. of angles, must be in [-PI,PI]:
 	      q = PhiP1 - Phi(iev,ifish,it)
101	      IF(q.LT.-PI) q = q + 2.D0*PI
	      IF(q.GT. PI) q = q - 2.D0*PI
	      IF(DABS(q).GT.PI) GOTO 101
	      DPhi(iev,ifish,it) = q
	    ENDDO	! end of this instant of time --> next it
	  ENDDO		! end of this fish --> next ifish
C
	  !Collective behavior observables:
	  DO ifish = 1,Nfish
	    DO it = itini,itend
	      DO jfish = 1,Nfish
		!Distance between ifish and jfish at time it of iev:
		q = DSQRT( (xx(iev,ifish,it) - xx(iev,jfish,it) )**2
     &			 + (yy(iev,ifish,it) - yy(iev,jfish,it) )**2)
		dij(iev,ifish,it,jfish) = max(q,1.D-4)
		!Angle with which ifish perceives jfish:
		!	psi_{ij} = theta_{ij} - \phi_i
		!where theta_{ij} = atan2(y_j - y_i, x_j - x_i) is
		!the angle that the vector FiFj forms with the velocity
		!vector of ifish: (Thetaij = Thetaji + PI, in [-pi,pi])
		Thetaij(iev,ifish,it,jfish) =
     &			ATAN2( yy(iev,jfish,it) -  yy(iev,ifish,it)
     &			      ,xx(iev,jfish,it) -  xx(iev,ifish,it))
		q = Thetaij(iev,ifish,it,jfish) - Phi(iev,ifish,it)
		psiij(iev,ifish,it,jfish) = q
		IF(q.LT.-PI) psiij(iev,ifish,it,jfish) = q + 2.D0*PI
		IF(q.GT. PI) psiij(iev,ifish,it,jfish) = q - 2.D0*PI
	      ENDDO
	    ENDDO
	  ENDDO ! End of this  fish --> next ifish,
	ENDDO 	! End of this event --> next  iev.
C	
C *********************************************************************
C 3) Perform reconstruction:
C	3.1) Fill the boxes resulting from the discretization
C	3.2) Calcualte the average of DPhi in each box, and the number
C	     of data in each box; this is stored in PP and eps resp.
C	     This part offers the possibility of using different
C	     cases for the origin of the data (see IOP in preamble)
C	3.3) Perform the iterative process until convergence
C	     This is done in subroutine ITERS, which also represents
C	     the result in output files 02.fR-end.dat, 02.gR-end.dat.

C *********************************************************************
	WRITE(*,*) 'Filling the boxes:'
	NN = Nd*Npsi
	WRITE(*,*) '---> #boxes: Nd*Npsi = ',NN
C Initialize PP and eps:
	eps = 0    !vector
	PP  = 0.D0 !vector
C
C 3.1) In which (id,ipsi)-box is the (iev,ifish,it)-th data?
C **** How many data are they in this box?
	ijoumeni = 0
	DO iev=1,Nevent		 ! **** ONLY ifish=1 MAKES SENSE **** !
	  DO ifish=1,1!Nfish	 ! the kicking instants of other fish !
	    DO it=1,Nt(iev)	 ! correspond to other events.        !
		ijoumeni = ijoumeni + 1
	      q4 =   dij(iev,ifish,it,3-ifish)
	      q5 = psiij(iev,ifish,it,3-ifish)
c
	      IF(    q4.GT.dijmin  .AND.q4.LT.dijmax
     &	        .AND.q5.GT.psiijmin.AND.q5.LT.psiijmax) THEN
		q1 =   dijmax -   dijmin + 1.D-14
		q2 = psiijmax - psiijmin
		id   = INT((q4 -   dijmin)*  Nd/q1) + 1
		ipsi = INT((q5 - psiijmin)*Npsi/q2) + 1
c
	        IF(ipsi.LT.1.OR.ipsi.GT.Npsi.OR.id.LT.1.OR.id.GT.Nd)
     &	        THEN
		  WRITE(*,'(" Bad index:",6I4)') id,ipsi
		  WRITE(*,'("Max values:",6I4)') Nd,Npsi
		  WRITE(*,*) 'Type 0 to continue or Ctrl C to stop'
		  READ(*,*) q
		ENDIF
		!The (iev,ifish,it)-th data is in box (id,ipsi):
		eps(id,ipsi) = eps(id,ipsi) + 1
	         PP(id,ipsi) =  PP(id,ipsi) + DPhi(iev,ifish,it)
	      ELSE
		NoValeVel = NoValeVel + 1
	      ENDIF
	    ENDDO
	  ENDDO
	ENDDO
	WRITE(*,'(" Number of times = Nlines/Nfish = ",I6)') ijoumeni
C
C 3.2) Calculate the mean value of DPhi in each (id,ipsi)-box
C **** and store it in PP: Note that some boxes are empty.
	ivide = 0
	epstot = 0
	DO id  =1,Nd
	DO ipsi=1,Npsi
	  IF(eps(id,ipsi).GT.0) THEN	! the box is not empty
	         epstot = epstot + eps(id,ipsi)
	    PP(id,ipsi) = PP(id,ipsi)/eps(id,ipsi)
	  ELSE			! the box is empty
	    ivide = ivide + 1
	  ENDIF  
	ENDDO
	ENDDO
C
C * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
C---[ To use analytical expressions (see preamble):
	IF(IOP.LT.3) THEN ! Data are produced by analyical expressions
	  PP = 0.D0 !vector
	  epstot = 0
	  DO id  =1,Nd
	  DO ipsi=1,Npsi
	    q1 = (id-0.5D0)*(dijmax - dijmin)/Nd + dijmin
	    q2 = (ipsi-0.5D0)*2.D0*PI/Npsi - PI
	    ndats = eps(id,ipsi)
	    IF(IOP.EQ.1) ndats = 1
	    DO l=1,ndats !Calculate average over ndats values
	      epstot = epstot + 1
	      PP(id,ipsi) = PP(id,ipsi) 
	!Comment or uncomment the following groups of lines
	!to select the function and the level of noise:
     &		+ fRepu(q1,RArena)*gRepu(q2)
     &	       	+ 0.7D0*(rand(0)-0.5D0) !noise
	    ! To check the hypothesis of separability:
c     &		+ 0.3D0*(4.D0*q1-DSIN(2.D0*q2))
c     &		+ 0.7D0*(rand(0)-0.5D0) !noise
	    ! A realistic one:
c     &		+ fRepu(q1,RArena)*gRepu(q2)
c     &		+ 1.13D0*DEXP( - (2.D0*q1/(3.D0*0.1D0))**2 )
c     &		* DSIN(q2) * DCOS(2.D0*q2)
c     &	      	+ 0.7D0*(rand(0)-0.5D0) !noise
	    ENDDO
	    IF(eps(id,ipsi).GT.0) PP(id,ipsi)=PP(id,ipsi)/ndats !mean
C
	    IF(IOP.EQ.1) THEN ! there is one and only one value per box
	      ivide = 0
	      NoValeVel = 0
	    ENDIF
	  ENDDO
	  ENDDO
	  IF(IOP.EQ.1) eps = 1 !data homogeneously distributed
	ENDIF  
C-------------]
C * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
C
	WRITE(*,'(" Boxes filled (empty): ",I6," (",I6,"), over ",I7)')
     &	  Nd*Npsi-ivide, ivide, Nd*Npsi
	WRITE(*,'("  Data placed  (lost): ",I6," (",I6,"), over ",I7)')
     &		epstot,NoValeVel,Nlines/Nfish
C
C *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
C 3.3) PERFORM THE ITERATIVE PROCESS UNTIL CONVERGENCE: *
	CALL ITERS(Nd,Npsi,PP,eps,dijmin,dijmax,fR,gR) !*
C *-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*-*
C
C *********************************************************************
C 4) Write the analytical functions used to build the data:
C *********************************************************************
	WRITE(*,*) 'Write the data from the analytical expressions'
	OPEN(702,file='00.fR-analytic.dat')
	OPEN(703,file='00.gR-analytic.dat')
	DO id=1,101
	  xm = (id-0.5D0)*2.D0*RArena/101.D0
	  WRITE(702,*) xm,fRepu(xm,RArena)
	ENDDO
	DO ipsi=1,101
	  xm = (ipsi-0.5D0)*2.D0*PI/101.D0 - PI
	  WRITE(703,*) xm,gRepu(xm)
	ENDDO
	CLOSE(702)
	CLOSE(703)
C
999	FORMAT(I6,2x,F10.4,1x,I7,5(1x,F10.4))
9999	FORMAT(A1,1x,I5,A15)
	END
C
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C Interaction functions: analytical expressions to check simple cases
C *********************************************************************
	FUNCTION fRepu(d,RArena)
	REAL*8   fRepu,d,RArena, gammaAtt, datt, latt
	gammaAtt = 4.D0
	datt = 0.03D0
	latt = 0.2D0
	fRepu = gammaAtt*(d-datt)/(1.D0+(d/latt)**2) ! [Calovi2018]
	IF (d.GT.0.35D0) fRepu = 0.D0
	
	!For the simple case:
	fRepu = dexp(-(d/0.1D0)**2)
	
	RETURN
	END
C ---------------------------
	FUNCTION gRepu(psi) ! Note: this function must be normalized
	REAL*8   gRepu,psi,x, a0,a1,a2
	x = psi
	gRepu = 1.395D0*DSIN(x)*(1.D0 - 0.33D0*DCOS(x)) ! [Calovi2018]
	
	!For the simple case:
	gRepu = DSQRT(2.D0)*DSIN(x)
	
	!Other realistic examples: (from [Calovi et al. (2018)]
c	gRepu = 0.93263D0*(1.D0 - 0.48D0*DCOS(x) - 0.31D0*DCOS(2.D0*x))
c	gRepu = 0.90123D0*(1.D0 + 0.6D0*DCOS(x) - 0.32D0*DCOS(2.D0*x))
c	gRepu = 1.6385D0*DSIN(x)*(1.D0 + 0.3D0*DCOS(2.D0*x))
	RETURN
	END
C
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C *********************************************************************
C SUBROUTINE PERFORMING THE ITERATIVE PROCESS
C    Inputs:    Nd, Npsi: number of nodes of functions fR, gR resp.
C		PP: average of the heading change in each (id,ipsi)-box
C		eps: number of data in each (id,ipsi)-box
C		dijmin,dijmax: range of values of state variable d_ij
C
C    Outputs:  	fR, gR: vectors of length Nd and Npsi resp. containing
C			the result of the iterative process.
C
C    Files generated:
C		01.fR-iters.dat: (3D) evolution of the iterations
C		02.fR-end.dat: (2D) final state at end of convergence.
C *********************************************************************
C
	SUBROUTINE ITERS(Nd,Npsi,PP,eps,dijmin,dijmax,fR,gR)
	REAL*8 PP(Nd,Npsi), MR(Nd,Npsi)
	INTEGER eps(Nd,Npsi), epstot, eeps
	REAL*8 fR(Nd),gR(Npsi), fR0(Nd),gR0(Npsi)
C
	REAL*8 dij(Nd),psiij(Npsi), dijmin,dijmax, fghR
	REAL*8 fRepu,gRepu
C
	REAL*8 lambda, TOL, PI, x, xm,  q,q1,q2,q3,q4
	REAL*8 sumfR, sumgR, NumfR, NumgR, DenomfR, DenomgR 
C
	REAL*8 DELTA, DELTAOld
	REAL*8 errfR,errgR
C	
	PI = DACOS(-1.D0)
	lambda = 0.025D0 ! Relaxation factor of the iterative process
	TOL = 1.D-9	 ! Tolerance for convergence
	Nit = 50000	 ! Maximum number of iteration
C Initial values:
	fR = 0.D0
	gR = 0.D0
	DELTA = 0.D0
C
C ---------------------------------------------------------------------
C Discretization: box centers
	DO id=1,Nd
	  dij(id) = (id-0.5D0)*(dijmax - dijmin)/Nd + dijmin
	ENDDO
	DO ipsi=1,Npsi
	  psiij(ipsi) = (ipsi-0.5D0)*2.D0*PI/Npsi - PI
	ENDDO
C ---------------------------------------------------------------------
C Initial condition of iterations:
	DO id=1,Nd
	  fR0(id) = 1.D0
	ENDDO
	DO ipsi=1,Npsi
	  gR0(ipsi) = 1.D0!DSQRT(2.D0)*DSIN(psiij(ipsi))
	ENDDO
C Representation of the initial values of the iterative process:
	OPEN(202,file='01.fR-iters.dat')
	OPEN(203,file='01.gR-iters.dat')
	DO id=1,Nd
	  WRITE(202,'(I7,3(2X,F18.6))') 0,dij(id),fR0(id)
	ENDDO  
	WRITE(202,*) !blank line for 3D-plot with gnuplot
C
	DO ipsi=1,Npsi
	  WRITE(203,'(I7,3(2X,F18.6))') 0,psiij(ipsi),gR0(ipsi)
	ENDDO  
	WRITE(203,*) !blank line for 3D-plot with gnuplot
C
C o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
C			     ITERATIONS
C o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o-o
C
	WRITE(*,*)
	WRITE(*,*)'-R E C O N S T R U C T I O N   I T E R A T I O N S-'
	WRITE(*,'(A26,1x,E8.2)') '  DELTA minimization: TOL=',TOL
C
	DO 1 it=1,Nit
C Common term in the expressions of f(d) and g(psi): MR
	  DO id  =1,Nd
	  DO ipsi=1,Npsi
	    MR(id,ipsi) = PP(id,ipsi)
	  ENDDO
	  ENDDO
C new f(d):
	  DO id=1,Nd
	      NumfR = 0.D0
	    DenomfR = 0.D0
	    DO ipsi=1,Npsi
	        NumfR =   NumfR + eps(id,ipsi) * gR0(ipsi)*MR(id,ipsi)
	      DenomfR = DenomfR + eps(id,ipsi) * gR0(ipsi)**2
	    ENDDO
	    fR(id) = fR0(id)
	    IF(DenomfR.GT.0.D0) fR(id) = NumfR/DenomfR
	  ENDDO
C new g(psi):
	  DO ipsi=1,Npsi
	      NumgR = 0.D0
	    DenomgR = 0.D0
	    DO id  =1,Nd
	        NumgR =   NumgR + eps(id,ipsi) * fR0(id)*MR(id,ipsi)
	      DenomgR = DenomgR + eps(id,ipsi) * fR0(id)**2
	    ENDDO
	    gR(ipsi) = gR0(ipsi)
	    IF(DenomgR.GT.0.D0) gR(ipsi) = NumgR/DenomgR
	  ENDDO
C
C Relaxed iteration:
	  fR0 = (1.D0-lambda)*fR0 + lambda*fR !vector
	  gR0 = (1.D0-lambda)*gR0 + lambda*gR !vector
C
C Normalization of angular function g(psi):
C   (1/2pi) int_{-pi}^{pi} g^2(s) ds = 1, so "*2PI/2PI" is omited
	  sumgR = 0.D0
	  DO ipsi=1,Npsi
	    sumgR = sumgR + gR0(ipsi)**2/Npsi
	  ENDDO
C Check integral:
	  IF(sumgR.GT.1.D-4) THEN
	    gR0 = gR0/DSQRT(sumgR) !vector
	    fR0 = fR0*DSQRT(sumgR) !vector
	  ELSE
	    WRITE(*,*)'problem: integral of g^2 is zero'
	    WRITE(*,'(5(F5.2,1x))') sumgR
	    WRITE(*,*)'Type Ctrl C to stop'
	    READ(*,*) q
	  ENDIF
C
C ---------------------------------------------------------------------
C Evolution of iterations (each 3 iterations):
	  IF(it/3*3.EQ.it) THEN
	    DO id=1,Nd
	      WRITE(202,'(I7,3(2X,F18.6))') it,dij(id),fR0(id)
	    ENDDO  
	    WRITE(202,*)  !blank line for 3D-plot with gnuplot
	    DO ipsi=1,Npsi
	      WRITE(203,'(I7,3(2X,F18.6))') it,psiij(ipsi),gR0(ipsi)
	    ENDDO  
	    WRITE(203,*)  !blank line for 3D-plot with gnuplot
C
C Evolution of DELTA = \sum_i \sum_j \sum_k eps_ijk [(...)^2+(...)^2]:
	    DELTAold = DELTA
	    DELTA = 0.D0
	    DO id  =1,Nd
	      DO ipsi=1,Npsi
	        q = PP(id,ipsi) - fR0(id)*gR0(ipsi)
	        DELTA = DELTA + eps(id,ipsi)*q**2
	      ENDDO
	    ENDDO
C
	    NN = Nd*Npsi
	    IF(DABS(DELTA - DELTAold).LT.TOL*NN) GOTO 123
C
	    WRITE(*,99, ADVANCE='NO') CHAR(13),it,' DELTA = ',DELTA/NN
	  ENDIF ! of representation
1	CONTINUE ! next iteration
C
C o-o-o-o-o-o-o-o-o-o-o-o-o-o END  ITERATIONS o-o-o-o-o-o-o-o-o-o-o-o-o
C
123	CLOSE(202)
	CLOSE(203)
C Put the values to which we have converged into the main variables:
	fR = fR0
	gR = gR0
C Check signs:  f x g = (-f) x (-g)
c	IF(fR(5).LT.0.D0) THEN
c	  fR = - fR !vector
c	  gR = - gR !vector
c	ENDIF  
C
C Represent final values:
	WRITE(*,*)
	OPEN(302,file='02.fR-end.dat')
	OPEN(303,file='02.gR-end.dat')
	    DO id=1,Nd
	      WRITE(302,'(3(2X,F18.6))') dij(id),fR(id)
	    ENDDO  
C
	    DO ipsi=1,Npsi
	      WRITE(303,'(3(2X,F18.6))') psiij(ipsi),gR(ipsi)
	    ENDDO  
	CLOSE(302)
	CLOSE(303)
C	  
C ---------------------------------------------------------------------
C Error estimates: (when possible, i.e., when data are produced by
C		    a model or directly with analytical expressions)
	WRITE(*,*) 'Errors: l_2 norm of relative error'
	WRITE(*,*) "[1/N sum_{i=1}^N e_i^2 ]^(1/2), e_i = x_i/x'_i - 1"
C fR:	
	OPEN(301,file='02.fR-end.dat')
	errfR = 0.D0
	DO id=1,Nd
	  READ(301,*) q,q2
	  IF(DABS(q2).GT.1.D-1)
     &		  errfR = errfR + (fRepu(q,0.25D0)/q2-1.D0)**2
	ENDDO
	errfR = DSQRT(errfR/Nd)
	WRITE(*,'(A8,F12.6)') '   f(d):',errfR
	CLOSE(301)
C gR:
	OPEN(301,file='02.gR-end.dat')
	errgR = 0.D0
	DO ipsi=1,Npsi
	  READ(301,*) q,q2
	  IF(DABS(q2).GT.1.D-1) errgR = errgR+(gRepu(q)/q2-1.D0)**2
	ENDDO
	errgR = DSQRT(errgR/Npsi)
	WRITE(*,'(A8,F12.6)') ' g(psi):',errgR
	CLOSE(301)
C ---------------------------------------------------------------------
99	FORMAT(A1,I6,1x,A8,3(1x,F18.12),1x,A9,6(1x,F5.2))	
	RETURN
	END
C
C End of file.